################################################################################
## Imports
################################################################################

from extronlib.interface import SerialInterface, EthernetClientInterface
from extronlib.system import Wait, Timer

import re

################################################################################
## Printing
################################################################################

def DebugPrint(method, data=None):
  print('WyreStorm.{0}[{1}]'.format(method, data if data is not None else ''))

################################################################################
## General Helpers
################################################################################

def FindKeyByValue(dictionary, value):
  for key, val in dictionary.items():
    if val.upper() == value.upper():
      return key

  return None

################################################################################
## Constants
################################################################################

MAX_STRIKES         = 3

PRIORITY_LEVELS     = 2
PRIORITY_COMMAND    = 0
PRIORITY_QUERY      = 1

COMMAND_TERMINATOR  = "\r\n"
RESPONSE_TERMINATOR = "\r\n"

################################################################################
## Mapping
################################################################################

MIN_INPUT_VOLUME     = -20
MAX_INPUT_VOLUME     = 20  
MAX_INPUT_VOLUME_MIC = 40

MIN_OUTPUT_VOLUME    = -100
MAX_OUTPUT_VOLUME    = 0

USB_HOST_MAP = {
  "Type C": "typec",
  "USB Host 1": "usbhost1",
  "USB Host 2": "usbhost2",
  "In 9": "in9",
  "In 10": "in10"
}

MODEL_INFO = {
  'MX-1007-HYB': {
    "Inputs": 10,
    "Outputs": 7,
    "Mics": 2,
    "Relays": 2,
    "OutputMap": {
      "Out 1": "out1",
      "Out 2": "out2",
      "Out 3": "out3",
      "Out 4": "out4",
      "Out 5": "out5",
      "Out 6": "out6",
      "Out 7": "out7",
      "All": "all"
    },
    "InputMap": {
      "In 1": "in1",
      "In 2": "in2",
      "In 3": "in3",
      "In 4": "in4",
      "In 5": "in5",
      "In 6": "in6",
      "In 7": "in7",
      "In 8": "in8",
      "In 9": "in9",
      "In 10": "in10",
      "None": "in0"
    },
    "AudioMixerInputMixMap": {
      "Mix 1": "1",
      "Mix 2": "2",
      "USB Mixer": "3"
    },
    "AudioMixerInputSourceMap": {
      "Out 1": "OUT1",
      "Out 2": "OUT2",
      "Out 3": "OUT3",
      "Out 4": "OUT4",
      "Out 5": "OUT5",
      "Out 6": "OUT6",
      "Out 7": "OUT7",
      "Dante In 1": "Dantein1",
      "Dante In 2": "Dantein2",
      "Dante In 3": "Dantein3",
      "Dante In 4": "Dantein4",
      "USB Host": "usbhost",
      "Mic In 1": "micin1",
      "Mic In 2": "micin2",
      "Line In": "linein"
    },
    "AudioMixerOutputMixMap": {
      "Mix 1": "1",
      "Mix 2": "2",
      "USB Mixer": "3"
    },
    "AudioMixerOutputSourceMap": {
      "Line Out 1": "LINEOUT1",
      "Line Out 2": "LINEOUT2",
      "Amp Out": "AMPOUT",
      "Dante Out 1/2": "DANTEOUT12",
      "Dante Out 3/4": "DANTEOUT34",
      "NHD Out": "NHDOUT",
    },
  },
  'MX-0804-EDC': {
    "Inputs": 8,
    "Outputs": 4,
    "Mics": 2,
    "Relays": 2,
    "OutputMap": {
      "Out 1": "out1",
      "Out 2": "out2",
      "Out 3": "out3",
      "Out 4": "out4",
      "All": "all"
    },
    "InputMap": {
      "In 1": "in1",
      "In 2": "in2",
      "In 3": "in3",
      "In 4": "in4",
      "In 5": "in5",
      "In 6": "in6",
      "In 7": "in7",
      "In 8": "in8",
      "None": "in0"
    },
    "AudioMixerInputMixMap": {
      "Mix 1": "1",
      "Mix 2": "2",
      "USB Mixer": "3"
    },
    "AudioMixerInputSourceMap": {
      "Out 1": "OUT1",
      "Out 2": "OUT2",
      "Out 3": "OUT3",
      "Out 4": "OUT4",
      "Dante In 1": "Dantein1",
      "Dante In 2": "Dantein2",
      "Dante In 3": "Dantein3",
      "Dante In 4": "Dantein4",
      "USB Host": "usbhost",
      "Mic In 1": "micin1",
      "Mic In 2": "micin2",
      "Line In": "linein"
    },
    "AudioMixerOutputMixMap": {
      "Mix 1": "1",
      "Mix 2": "2",
      "USB Mixer": "3"
    },
    "AudioMixerOutputSourceMap": {
      "Line Out 1": "LINEOUT1",
      "Line Out 2": "LINEOUT2",
      "Amp Out": "AMPOUT",
      "Dante Out 1/2": "DANTEOUT12",
      "Dante Out 3/4": "DANTEOUT34",
      "NHD Out": "NHDOUT",
    },
  }
}

################################################################################
## Device Class
################################################################################

class DeviceClass:
  def __init__(self, Model):
    self.models = ['MX-1007-HYB', 'MX-0804-EDC']
    self.info = MODEL_INFO.get(Model)

    if self.info:
      self.inputCount = self.info["Inputs"]
      self.outputCount = self.info["Outputs"]
      self.micCount = self.info["Mics"]
      self.relayCount = self.info["Relays"]
      self.enableDebug = False
      self.enableSSL = False
      self.initializing = False
      self.debug = False
      self.queue = PriorityQueue(PRIORITY_LEVELS)
      self.queueLocked = False
      self.lastOut = None
      self.responseTimer = None
      self.responseTime = 5
      self.pacerTimer = None
      self.pacerTime = isinstance(self, EthernetClass) and .1 or .3 # RS-232 seems to struggle when messages are sent too quickly so we slow it down in that case
      self.strikeCount = 0
      self.heartbeatTimer = None
      self.heartbeatTime = 30
      self.buffer = b''
      self.parsing = False
      self.status = {
        'Communicating': CommunicatingStatus(),
        'Initialized': InitializedStatus(),
        'VideoRoute': VideoRouteStatus(self),
        'PhantomPower': PhantomPowerStatus(self),
        'USBHost': USBHostStatus(self),
        'RelayControl': RelayStatus(self),
        'AudioMixerInputSource': AudioMixerInputSourceStatus(self),
        'AudioMixerInputVolume': AudioMixerInputVolumeStatus(self),
        'AudioMixerInputMute': AudioMixerInputMuteStatus(self),
        'AudioMixerOutputSource': AudioMixerOutputSourceStatus(self),
        'AudioMixerOutputVolume': AudioMixerOutputVolumeStatus(self),
        'AudioMixerOutputMute': AudioMixerOutputMuteStatus(self),
      }

      self.Connected = self.__Connected
      self.ReceiveData = self.__ReceiveData

  # --------------------------------------------------------------------------
  # Printing Helpers
  # --------------------------------------------------------------------------

  def Debug(self, *args):
    if self.enableDebug:
      if len(args) == 1:
        DebugPrint(args[0])
      
      elif len(args) == 2:
        DebugPrint(args[0], data=args[1])

  # --------------------------------------------------------------------------
  # Connection/Communication
  # --------------------------------------------------------------------------

  def __Connected(self, interface, state):
    self.Debug("Connected")

    self.SendHeartbeat()


  def __ReceiveData(self, interface, data):
    self.HandleResponse(data)

  # --------------------------------------------------------------------------
  # Initialization Helpers
  # --------------------------------------------------------------------------
  
  def GetInitialized(self):
    if not self.initializing and not self.IsInitialized():
      self.initializing = True
    
      self.Debug('GetInitialized')
      
      for name, status in self.status.items():
        if not name in ['Communicating', 'Initialized']:
          status.GetInitialized()


  def IsInitialized(self):
    if not self.status['Communicating'].value:
      return False
    
    for name, status in self.status.items():
      if not name in ['Communicating', 'Initialized']:
        if not status.IsInitialized():
          return False
        
    return True  

  # --------------------------------------------------------------------------
  # Exposed Methods (Intended to be called directly by consumer of module)
  # --------------------------------------------------------------------------

  def Reinitialize(self):
    self.Debug("Reinitialize")

    self.Reset()

    if isinstance(self, EthernetClass):
      result = self.Connect()
      print("Connect result: {}".format(result))

      if result == "ConnectedAlready":
        self.SendHeartbeat()
    
    else:
      self.SendHeartbeat()


  def Set(self, command, value, qualifier=None):
    method = getattr(self, 'Set%s' % command, None)

    if method is not None and callable(method):
      method(value, qualifier)
    
    else:
      raise AttributeError("{0} is not a valid Set command.".format(command))


  def Update(self, command, qualifier=None):
    method = getattr(self, 'Update%s' % command, None)
    
    if method is not None and callable(method):
      method(None, qualifier)
    
    else:
      raise AttributeError("{0} is not a valid Update command.".format(command))
  

  def SubscribeStatus(self, command, qualifier, callback):
    if command in self.status:
      self.status[command].SubscribeStatus(callback, qualifier=qualifier)

    else:
      raise KeyError('{0} is not a valid SubscribeStatus command'.format(command))


  def ReadStatus(self, command, qualifier=None):
    if command in self.status:
        return self.status[command].ReadStatus(qualifier)

    else:
      raise KeyError('{0} is not a valid ReadStatus command'.format(command))
      return None

  # --------------------------------------------------------------------------
  # General Helpers
  # --------------------------------------------------------------------------

  def GetStateValue(self, value, qualifier, statusName):
    state = value == 'On' and 'on' or 'off'
    
    if value == 'Toggle':
      current = self.status[statusName].ReadStatus(qualifier)
    
      if current is None:
        return None

      state = current == 'Off' and 'on' or 'off'

    return state


  def GetAudioMixerInputInfo(self, qualifier):
    return self.GetAudioMixerInputMixName(qualifier), self.GetAudioMixerInputSourceName(qualifier)


  def GetAudioMixerInputMixName(self, qualifier):
    mix_mapped = None

    mix = qualifier and qualifier.get('Mix') or None

    if mix:
      mix_mapped = self.info["AudioMixerInputMixMap"].get(mix)

    return mix_mapped


  def GetAudioMixerInputSourceName(self, qualifier):
    source_mapped = None

    source = qualifier and qualifier.get('Source') or None

    if source:
      source_mapped = self.info["AudioMixerInputSourceMap"].get(source)

    return source_mapped

  
  def GetAudioMixerOutputInfo(self, qualifier):
    return self.GetAudioMixerOutputMixName(qualifier), self.GetAudioMixerOutputSourceName(qualifier)


  def GetAudioMixerOutputMixName(self, qualifier):
    mix_mapped = None
    
    mix = qualifier and qualifier.get('Mix') or None

    if mix:
      mix_mapped = self.info["AudioMixerOutputMixMap"].get(mix)

    return mix_mapped


  def GetAudioMixerOutputSourceName(self, qualifier):
    source_mapped = None
    
    source = qualifier and qualifier.get('Source') or None

    if source:
      source_mapped = self.info["AudioMixerOutputSourceMap"].get(source)

    return source_mapped

  # --------------------------------------------------------------------------
  # Set Helpers
  # --------------------------------------------------------------------------

  def SetVideoRoute(self, value, qualifier):
    self.Debug("SetVideoRoute", "{} | {}".format(value, qualifier))

    output = qualifier and qualifier.get('Output') or None

    if value and output:
      input_mapped = self.info["InputMap"].get(value)
      output_mapped = self.info["OutputMap"].get(output)

      if input_mapped and output_mapped:
        request = CommandBuilder.SetVideoRoute(input_mapped, output_mapped)

        if request:
          self.Enqueue(request, PRIORITY_COMMAND)

  
  def SetAudioMixerInputSource(self, value, qualifier):
    self.Debug("SetAudioMixerInputSource", "{} | {}".format(value, qualifier))

    (mix_mapped, source_mapped) = self.GetAudioMixerInputInfo(qualifier)

    if mix_mapped and source_mapped:
      request = CommandBuilder.SetAudioMixerInputSource(int(mix_mapped), source_mapped, self.GetStateValue(value, qualifier, 'AudioMixerInputSource'))

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)

  
  def SetAudioMixerInputVolume(self, value, qualifier):
    self.Debug("SetAudioMixerInputVolume", "{} | {}".format(value, qualifier))

    (mix_mapped, source_mapped) = self.GetAudioMixerInputInfo(qualifier)

    if mix_mapped and source_mapped:
      max_ = (source_mapped == "micin1" or source_mapped == "micin2") and MAX_INPUT_VOLUME_MIC or MAX_INPUT_VOLUME
      val = int(value)

      if val < MIN_INPUT_VOLUME or val > max_:
        return
      
      request = CommandBuilder.SetAudioMixerInputVolume(int(mix_mapped), source_mapped, val)

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)


  def SetAudioMixerInputMute(self, value, qualifier):
    self.Debug("SetAudioMixerInputMute", "{} | {}".format(value, qualifier))

    (mix_mapped, source_mapped) = self.GetAudioMixerInputInfo(qualifier)

    if mix_mapped and source_mapped:
      request = CommandBuilder.SetAudioMixerInputMute(int(mix_mapped), source_mapped, self.GetStateValue(value, qualifier, 'AudioMixerInputMute'))

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)


  def SetAudioMixerOutputSource(self, value, qualifier):
    self.Debug("SetAudioMixerOutputSource", "{} | {}".format(value, qualifier))

    (mix_mapped, source_mapped) = self.GetAudioMixerOutputInfo(qualifier)

    if mix_mapped and source_mapped:
      request = CommandBuilder.SetAudioMixerOutputSource(int(mix_mapped), source_mapped, self.GetStateValue(value, qualifier, 'AudioMixerOutputSource'))

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)

  
  def SetAudioMixerOutputVolume(self, value, qualifier):
    self.Debug("SetAudioMixerOutputVolume", "{} | {}".format(value, qualifier))

    mix_mapped = self.GetAudioMixerOutputMixName(qualifier)

    if mix_mapped:
      val = int(value)

      if val < MIN_OUTPUT_VOLUME or val > MAX_OUTPUT_VOLUME:
        return
      
      request = CommandBuilder.SetAudioMixerOutputVolume(int(mix_mapped), val)

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)


  def SetAudioMixerOutputMute(self, value, qualifier):
    self.Debug("SetAudioMixerOutputMute", "{} | {}".format(value, qualifier))

    mix_mapped = self.GetAudioMixerOutputMixName(qualifier)

    if mix_mapped:
      request = CommandBuilder.SetAudioMixerOutputMute(int(mix_mapped), self.GetStateValue(value, qualifier, 'AudioMixerOutputMute'))

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)

  
  def SetDisplayPower(self, value, qualifier):
    self.Debug("SetDisplayPower", "{} | {}".format(value, qualifier))

    output = qualifier and qualifier.get('Display') or None

    if output:
      output_mapped = self.info["OutputMap"].get(output)

      if output_mapped:
        state = value == 'On' and 'on' or 'off'

        request = CommandBuilder.SetDisplayPower(output_mapped, state)

        if request:
          self.Enqueue(request, PRIORITY_COMMAND)

  
  def SetPhantomPower(self, value, qualifier):
    self.Debug("SetPhantomPower", "{} | {}".format(value, qualifier))

    mic = qualifier and qualifier.get('Mic') or None

    if mic and mic > 0 and mic <= self.micCount:
      request = CommandBuilder.SetPhantomPower(mic, self.GetStateValue(value, qualifier, 'PhantomPower'))

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)


  def SetUSBHost(self, value, qualifier):
    self.Debug("SetUSBHost", "{} | {}".format(value, qualifier))   

    if value:
      host_mapped = USB_HOST_MAP.get(value)

      if host_mapped:
        request = CommandBuilder.SetUSBHost(host_mapped)

        if request:
          self.Enqueue(request, PRIORITY_COMMAND) 
  

  def SetRelay(self, value, qualifier):
    self.Debug("SetRelay", "{} | {}".format(value, qualifier))

    relay = qualifier and qualifier.get('Relay') or None

    if relay and relay > 0 and relay <= self.relayCount:
      request = CommandBuilder.SetRelayState(relay, self.GetStateValue(value, qualifier, 'RelayControl'))

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)
      

  def SetScene(self, value, qualifier):
    self.Debug("SetScene", "{} | {}".format(value, qualifier))

    scene = qualifier and qualifier.get('Scene') or None

    if scene:
      request = None

      if value == 'Save':
        request = CommandBuilder.SetSceneSave(scene)

      elif value == 'Recall':
        request = CommandBuilder.SetSceneRecall(scene)

      if request:
        self.Enqueue(request, PRIORITY_COMMAND)

  # --------------------------------------------------------------------------
  # Update Helpers
  # --------------------------------------------------------------------------

  def UpdateVideoRoute(self, value, qualifier):
    self.Debug("UpdateVideoRoute", "{}".format(qualifier))

    output = qualifier and qualifier.get('Output') or None

    if output:
      output_mapped = self.info["OutputMap"].get(output)

      if output_mapped:
        request = CommandBuilder.GetVideoRoute(output_mapped)

        if request:
          self.Enqueue(request, PRIORITY_QUERY)


  def UpdateAudioMixerInputSource(self, value, qualifier):
    self.Debug("UpdateAudioMixerInputSource", "{} | {}".format(value, qualifier))
    
    mix_mapped = self.GetAudioMixerInputMixName(qualifier)

    if mix_mapped:
      request = CommandBuilder.GetAudioMixerInputSource(int(mix_mapped))

      if request:
        self.Enqueue(request, PRIORITY_QUERY)

  
  def UpdateAudioMixerInputVolume(self, value, qualifier):
    self.Debug("UpdateAudioMixerInputVolume", "{} | {}".format(value, qualifier))
    
    (mix_mapped, source_mapped) = self.GetAudioMixerInputInfo(qualifier)

    if mix_mapped and source_mapped:
      request = CommandBuilder.GetAudioMixerInputVolume(int(mix_mapped), source_mapped)

      if request:
        self.Enqueue(request, PRIORITY_QUERY)


  def UpdateAudioMixerInputMute(self, value, qualifier):
    self.Debug("UpdateAudioMixerInputMute", "{} | {}".format(value, qualifier))
    
    (mix_mapped, source_mapped) = self.GetAudioMixerInputInfo(qualifier)

    if mix_mapped and source_mapped:
      request = CommandBuilder.GetAudioMixerInputMute(int(mix_mapped), source_mapped)

      if request:
        self.Enqueue(request, PRIORITY_QUERY)

  
  def UpdateAudioMixerOutputSource(self, value, qualifier):
    self.Debug("UpdateAudioMixerOutputSource", "{} | {}".format(value, qualifier))
    
    mix_mapped = self.GetAudioMixerOutputMixName(qualifier)

    if mix_mapped:
      request = CommandBuilder.GetAudioMixerOutputSource(int(mix_mapped))

      if request:
        self.Enqueue(request, PRIORITY_QUERY)


  def UpdateAudioMixerOutputVolume(self, value, qualifier):
    self.Debug("UpdateAudioMixerOutputVolume", "{} | {}".format(value, qualifier))
    
    mix_mapped = self.GetAudioMixerOutputMixName(qualifier)

    if mix_mapped:
      request = CommandBuilder.GetAudioMixerOutputVolume(int(mix_mapped))

      if request:
        self.Enqueue(request, PRIORITY_QUERY)
    
  
  def UpdateAudioMixerOutputMute(self, value, qualifier):
    self.Debug("UpdateAudioMixerOutputMute", "{} | {}".format(value, qualifier))
    
    mix_mapped = self.GetAudioMixerOutputMixName(qualifier)

    if mix_mapped:
      request = CommandBuilder.GetAudioMixerOutputMute(int(mix_mapped))

      if request:
        self.Enqueue(request, PRIORITY_QUERY)


  def UpdatePhantomPower(self, value, qualifier):
    self.Debug("UpdatePhantomPower", "{} | {}".format(value, qualifier))

    mic = qualifier and qualifier.get('Mic') or None

    if mic and mic > 0 and mic <= self.micCount:
      request = CommandBuilder.GetPhantomPower(mic)

      if request:
        self.Enqueue(request, PRIORITY_QUERY)


  def UpdateUSBHost(self, value, qualifier):
    self.Debug("UpdateUSBHost", "{} | {}".format(value, qualifier))   

    request = CommandBuilder.GetUSBHost()

    if request:
      self.Enqueue(request, PRIORITY_QUERY) 
  

  def UpdateRelayControl(self, value, qualifier):
    self.Debug("UpdateRelayControl", "{} | {}".format(value, qualifier))

    relay = qualifier and qualifier.get('Relay') or None

    if relay and relay > 0 and relay <= self.relayCount:
      request = CommandBuilder.GetRelayState(relay)

      if request:
        self.Enqueue(request, PRIORITY_QUERY)

  # --------------------------------------------------------------------------
  # Heartbeat Helpers
  # --------------------------------------------------------------------------
  
  def SendHeartbeat(self):
    self.RestartHeartbeatTimer()

    if self.queue.IsEmpty():
      self.Debug('SendHeartbeat')

      request = CommandBuilder.GetFirmwareVersion()
      self.Enqueue(request, PRIORITY_QUERY)


  def StartHeartbeatTimer(self):
    self.heartbeatTimer = Wait(self.heartbeatTime, self.SendHeartbeat)


  def StopHeartbeatTimer(self):
    if self.heartbeatTimer is not None:
      self.heartbeatTimer.Cancel()
      self.heartbeatTimer = None


  def RestartHeartbeatTimer(self):
    self.StopHeartbeatTimer()
    self.StartHeartbeatTimer()

  # --------------------------------------------------------------------------
  # Command Helpers
  # --------------------------------------------------------------------------

  def SendMessage(self, request):
    if request:
      self.lastOut = request
      self.queueLocked = True

      self.Debug('SendMessage', request.rstrip(COMMAND_TERMINATOR))
      self.Send(request)

      self.RestartResponseTimer()


  def StartPacerTimer(self):
    self.pacerTimer = Wait(self.pacerTime, self.PacerTimerEvent)


  def StopPacerTimer(self):        
    if self.pacerTimer is not None:
      self.pacerTimer.Cancel()
      self.pacerTimer = None


  def RestartPacerTimer(self):
    self.StopPacerTimer()
    self.StartPacerTimer()


  def PacerTimerEvent(self):
    self.queueLocked = False
    self.Dequeue()

  # --------------------------------------------------------------------------
  # Queue Helpers
  # --------------------------------------------------------------------------

  def Enqueue(self, request, priority):
    self.queue.Enqueue(request, priority)

    if self.queue.Size() == 1:
      self.Dequeue()


  def Dequeue(self):
    if self.queue.Size() > 0 and not self.queueLocked:
      self.SendMessage(self.queue.Dequeue() + COMMAND_TERMINATOR)

  # --------------------------------------------------------------------------
  # Response Helpers
  # --------------------------------------------------------------------------

  def HandleResponse(self, response):
    try:
      self.buffer += response
      location = -1

      if not self.parsing:
        self.parsing = True

        while True:
          location = self.buffer.decode().find(RESPONSE_TERMINATOR)
          
          if location < 0:
            break

          end = location + len(RESPONSE_TERMINATOR)
          r = self.buffer[:end].decode()
          self.buffer = self.buffer[end:]

          self.StopResponseTimer()

          if not self.status['Communicating'].value:
            self.status['Communicating'].SetStatus(True)

          self.ProcessResponse(r)

        self.parsing = False

      self.lastOut = None
      
      if not self.status['Initialized'].value:
        if self.initializing and self.IsInitialized():
          self.initializing = False
          self.status['Initialized'].SetStatus(True)
      
      self.RestartPacerTimer()

    except Exception as e:
      self.Debug('HandleResponse.Exception', str(e))
      self.queue.Clear()
      self.queueLocked = False


  def ProcessResponse(self, response):
    try:
      self.Debug("ProcessResponse", response.rstrip(RESPONSE_TERMINATOR))

      patterns = {
        r'VER.*?(?=\r\n|$)': self.ProcessHeartbeatResponse,
        r'MP (IN\d+) (OUT\d+)(?=\r\n|$)': self.ProcessVideoRouteResponse,
        r'SW (IN\d+) (OUT\d+)(?=\r\n|$)': self.ProcessVideoRouteResponse,
        r'SW (IN\d+) ALL(?=\r\n|$)': self.ProcessVideoRouteAllResponse,
        r'MIC_PHA_PWR (\d+) (OFF|ON)(?=\r\n|$)': self.ProcessPhantomPowerResponse,
        r'USB_HOST ([a-zA-Z0-9]+)(?=\r\n|$)': self.ProcessUSBHostResponse,
        r'RELAY (\d+) (OFF|ON)(?=\r\n|$)': self.ProcessRelayResponse,
        r'MIXER_IN (\d+) (\w+) (ON|OFF)(?=\r\n|$)': self.ProcessAudioMixerInputSourceResponse,
        r'MIXER_IN_VOL (\d+) (\w+) (-?\d+(\.\d+)?)(?=\r\n|$)': self.ProcessAudioMixerInputVolumeResponse,
        r'MIXER_IN_MUTE (\d+) (\w+) (ON|OFF)(?=\r\n|$)': self.ProcessAudioMixerInputMuteResponse,
        r'MIXER_OUT (\d+) (\w+) (ON|OFF)(?=\r\n|$)': self.ProcessAudioMixerOutputSourceResponse,
        r'MIXER_OUT_VOL (\d+) (-?\d+(\.\d+)?)(?=\r\n|$)': self.ProcessAudioMixerOutputVolumeResponse,
        r'MIXER_OUT_MUTE (\d+) (ON|OFF)(?=\r\n|$)': self.ProcessAudioMixerOutputMuteResponse,
      }

      for pattern, function in patterns.items():
        matches = re.finditer(pattern, response)

        for match in matches:
          function(match, response)

    except Exception as e:
      self.Debug('ProcessResponse.Exception', str(e))


  def ProcessHeartbeatResponse(self, match, response):
    self.GetInitialized()


  def ProcessVideoRouteResponse(self, match, response):
    self.status["VideoRoute"].SetStatus(match.group(2), match.group(1))

  
  def ProcessVideoRouteAllResponse(self, match, response):
    for output in self.info["OutputMap"].keys():
      if output != "All":
        self.status["VideoRoute"].SetStatus(self.info["OutputMap"][output].upper(), match.group(1))


  def ProcessAudioMixerInputSourceResponse(self, match, response):
    self.status["AudioMixerInputSource"].SetStatus(match.group(1), match.group(2), match.group(3))


  def ProcessAudioMixerInputVolumeResponse(self, match, response):
    self.status["AudioMixerInputVolume"].SetStatus(match.group(1), match.group(2), match.group(3))


  def ProcessAudioMixerInputMuteResponse(self, match, response):
    self.status["AudioMixerInputMute"].SetStatus(match.group(1), match.group(2), match.group(3))


  def ProcessAudioMixerOutputSourceResponse(self, match, response):
    self.status["AudioMixerOutputSource"].SetStatus(match.group(1), match.group(2), match.group(3))


  def ProcessAudioMixerOutputVolumeResponse(self, match, response):
    self.status["AudioMixerOutputVolume"].SetStatus(match.group(1), match.group(2))


  def ProcessAudioMixerOutputMuteResponse(self, match, response):
    self.status["AudioMixerOutputMute"].SetStatus(match.group(1), match.group(2))


  def ProcessPhantomPowerResponse(self, match, response):
    self.status["PhantomPower"].SetStatus(int(match.group(1)), match.group(2))


  def ProcessUSBHostResponse(self, match, response):
    self.status["USBHost"].SetStatus(match.group(1))


  def ProcessRelayResponse(self, match, response):
    self.status["RelayControl"].SetStatus(int(match.group(1)), match.group(2))


  def StartResponseTimer(self):
    self.responseTimer = Wait(self.responseTime, self.ResponseTimerEvent)


  def StopResponseTimer(self):        
    if self.responseTimer is not None:
      self.responseTimer.Cancel()
      self.responseTimer = None


  def RestartResponseTimer(self):
    self.StopResponseTimer()
    self.StartResponseTimer()


  def ResponseTimerEvent(self):
    self.Debug('ResponseTimerEvent')
    
    self.strikeCount += 1
        
    if self.strikeCount == MAX_STRIKES:
      self.Strikeout()
        
    else:
      self.lastOut = None
      self.queueLocked = False
      self.Dequeue()
        
    if self.queue.IsEmpty():
      self.heartbeatTimer = Wait(3, self.SendHeartbeat)


  def Reset(self):
    self.Debug('Reset')

    self.status['Communicating'].SetStatus(False)
    self.status['Initialized'].SetStatus(False)

    for name, status in self.status.items():
      if not name in ['Communicating', 'Initialized']:
        status.Reset()
    
    self.initializing = False
    self.queue = PriorityQueue(PRIORITY_LEVELS)
    self.queueLocked = False
    self.lastOut = None
    self.strikeCount = 0
    self.buffer = b''
    self.parsing = False
    self.StopHeartbeatTimer()
    self.StopResponseTimer()
    self.StopPacerTimer()


  def Strikeout(self):
    self.Debug('Strikeout')

    self.Reinitialize()

################################################################################
## Command Builder
################################################################################

class CommandBuilder():
  def __init__(self):
    pass
  
  @staticmethod
  def GetFirmwareVersion():
    return "GET VER"

  @staticmethod
  def GetVideoRoute(output):
    return "GET MP {}".format(output)

  @staticmethod
  def SetVideoRoute(input, output):
    return "SET SW {} {}".format(input, output)

  @staticmethod
  def GetAudioMixerInputSource(mix):
    return "GET MIXER_IN {}".format(mix)

  @staticmethod
  def SetAudioMixerInputSource(mix, source, state):
    return "SET MIXER_IN {} {} {}".format(mix, source, state)

  @staticmethod
  def GetAudioMixerInputVolume(mix, source):
    return "GET MIXER_IN_VOL {} {}".format(mix, source)

  @staticmethod
  def SetAudioMixerInputVolume(mix, source, level):
    return "SET MIXER_IN_VOL {} {} {}".format(mix, source, level)

  @staticmethod
  def GetAudioMixerInputMute(mix, source):
    return "GET MIXER_IN_MUTE {} {}".format(mix, source)

  @staticmethod
  def SetAudioMixerInputMute(mix, source, state):
    return "SET MIXER_IN_MUTE {} {} {}".format(mix, source, state)

  @staticmethod
  def GetAudioMixerOutputSource(mix):
    return "GET MIXER_OUT {}".format(mix)

  @staticmethod
  def SetAudioMixerOutputSource(mix, source, state):
    return "SET MIXER_OUT {} {} {}".format(mix, source, state)

  @staticmethod
  def GetAudioMixerOutputVolume(mix):
    return "GET MIXER_OUT_VOL {}".format(mix)

  @staticmethod
  def SetAudioMixerOutputVolume(mix, level):
    return "SET MIXER_OUT_VOL {} {}".format(mix, level)

  @staticmethod
  def GetAudioMixerOutputMute(mix):
    return "GET MIXER_OUT_MUTE {}".format(mix)

  @staticmethod
  def SetAudioMixerOutputMute(mix, state):
    return "SET MIXER_OUT_MUTE {} {}".format(mix, state)

  @staticmethod
  def SetDisplayPower(output, state):
    return "SET CEC_PWR {} {}".format(output, state)

  @staticmethod
  def GetPhantomPower(mic):
    return "GET MIC_PHA_PWR {}".format(mic)

  @staticmethod
  def SetPhantomPower(mic, state):
    return "SET MIC_PHA_PWR {} {}".format(mic, state)

  @staticmethod
  def GetUSBHost():
    return "GET USB_HOST"

  @staticmethod
  def SetUSBHost(host):
    return "SET USB_HOST {}".format(host)

  @staticmethod
  def GetRelayState(relay):
    return "GET RELAY {}".format(relay)

  @staticmethod
  def SetRelayState(relay, state):
    return "SET RELAY {} {}".format(relay, state)

  @staticmethod
  def SetSceneSave(scene):
    return "SAVE PRESET {}".format(scene)

  @staticmethod
  def SetSceneRecall(scene):
    return "RESTORE PRESET {}".format(scene)

################################################################################
## Status
################################################################################

class Status():
  def __init__(self, name=None):
    self.name = name
    self.value = None
    self.subscribeCallback = None
    
  def ReadStatus(self, qualifier=None):
    return self.value
    
  def SubscribeStatus(self, callback, qualifier=None):
    if callback is not None and callable(callback):
      self.subscribeCallback = callback
  
  def SetStatus(self, value):
    self.value = value
    self.UpdateStatus()

  def UpdateStatus(self):
    if self.name is not None and self.subscribeCallback is not None:
      self.subscribeCallback(self.name, self.ReadStatus(), None)
      
  def IsInitialized(self):
    return self.value is not None


class CommunicatingStatus(Status):
  def __init__(self):
    super().__init__('Communicating')
    
  def ReadStatus(self, qualifier=None):
    return 'Communicating' if self.value else 'Not Communicating'


class InitializedStatus(Status):
  def __init__(self):
    super().__init__('Initialized')
    
  def ReadStatus(self, qualifier=None):
    return 'Initialized' if self.value else 'Not Initialized'


class VideoRouteStatus(Status):
  def __init__(self, listener):
    super().__init__('VideoRoute')

    self.listener = listener
    self.outputMap = listener.info["OutputMap"]
    self.inputMap = listener.info["InputMap"]
    
    self.outputs = {}

    for output in self.outputMap.keys():
      if output != "All":
        self.outputs[output] = None

  def Reset(self):
    for output in self.outputs.keys():
      self.outputs[output] = None
      self.UpdateStatus(output)

  def ReadStatus(self, qualifier):
    if "Output" in qualifier:
      if qualifier["Output"] in self.outputs:
        return self.outputs[qualifier["Output"]]
    
    return None
    
  def SetStatus(self, output, input):
    output_mapped = FindKeyByValue(self.outputMap, output)
    input_mapped  = FindKeyByValue(self.inputMap, input)

    if output_mapped and input_mapped:
      self.outputs[output_mapped] = input_mapped
      self.UpdateStatus(output_mapped)

  def UpdateStatus(self, output):
    if output in self.outputs and self.name is not None and self.subscribeCallback is not None:
      self.subscribeCallback(self.name, self.outputs[output], { "Output": output })

  def GetInitialized(self):
      request = CommandBuilder.GetVideoRoute("all")

      if request:
        self.listener.Enqueue(request, PRIORITY_QUERY)

  def IsInitialized(self):
    for output, value in self.outputs.items():
      if value is None:
        return False
    
    return True


class PhantomPowerStatus(Status):
  def __init__(self, listener):
    super().__init__('PhantomPower')

    self.listener = listener

    self.mics = [None] * listener.micCount

  def Reset(self):
    for i, state in enumerate(self.mics):
      state = None
      self.UpdateStatus(i + 1)

  def ReadStatus(self, qualifier):
    if "Mic" in qualifier:
      mic = qualifier["Mic"]
      
      if mic > 0 and mic <= self.listener.micCount:
        return self.mics[mic - 1]
    
    return None
    
  def SetStatus(self, mic, state):
    if mic > 0 and mic <= self.listener.micCount:
      self.mics[mic - 1] = state.title()
      self.UpdateStatus(mic)

  def UpdateStatus(self, mic):
    if mic > 0 and mic <= self.listener.micCount and self.name is not None and self.subscribeCallback is not None:
      self.subscribeCallback(self.name, self.mics[mic - 1], { "Mic": mic })

  def GetInitialized(self):
      for i, _ in enumerate(self.mics):
        request = CommandBuilder.GetPhantomPower(i + 1)

        if request:
          self.listener.Enqueue(request, PRIORITY_QUERY)

  def IsInitialized(self):
    for _, state in enumerate(self.mics):
      if state is None:
        return False
    
    return True


class USBHostStatus(Status):
  def __init__(self, listener):
    super().__init__('USBHost')

    self.listener = listener
    
    self.host = None

  def Reset(self):
    self.host = None
    self.UpdateStatus(None)

  def ReadStatus(self, qualifier):
    return self.host
    
  def SetStatus(self, host):
    host_mapped = FindKeyByValue(USB_HOST_MAP, host.lower())

    if host_mapped:
      self.host = host_mapped
      self.UpdateStatus(host_mapped)

  def UpdateStatus(self, host):
    if self.name is not None and self.subscribeCallback is not None:
      self.subscribeCallback(self.name, host, None)

  def GetInitialized(self):
      request = CommandBuilder.GetUSBHost()

      if request:
        self.listener.Enqueue(request, PRIORITY_QUERY)

  def IsInitialized(self):
    return self.host is not None


class RelayStatus(Status):
  def __init__(self, listener):
    super().__init__('RelayControl')

    self.listener = listener

    self.relays = [None] * listener.relayCount

  def Reset(self):
    for i, state in enumerate(self.relays):
      state = None
      self.UpdateStatus(i + 1)

  def ReadStatus(self, qualifier):
    if "Relay" in qualifier:
      relay = qualifier["Relay"]
      
      if relay > 0 and relay <= self.listener.relayCount:
        return self.relays[relay - 1]
    
    return None
    
  def SetStatus(self, relay, state):
    if relay > 0 and relay <= self.listener.relayCount:
      self.relays[relay - 1] = state.title()
      self.UpdateStatus(relay)

  def UpdateStatus(self, relay):
    if relay > 0 and relay <= self.listener.relayCount and self.name is not None and self.subscribeCallback is not None:
      self.subscribeCallback(self.name, self.relays[relay - 1], { "Relay": relay })

  def GetInitialized(self):
      for i, _ in enumerate(self.relays):
        request = CommandBuilder.GetRelayState(i + 1)

        if request:
          self.listener.Enqueue(request, PRIORITY_QUERY)

  def IsInitialized(self):
    for _, state in enumerate(self.relays):
      if state is None:
        return False
    
    return True


class AudioMixerInputStatus(Status):
  def __init__(self, listener, name):
    super().__init__(name)

    self.listener = listener
    self.mixMap = listener.info["AudioMixerInputMixMap"]
    self.sourceMap = listener.info["AudioMixerInputSourceMap"]
    
    self.mixes = {}

    for mix in self.mixMap.keys():
      self.mixes[mix] = {}
      for source in self.sourceMap.keys():
        self.mixes[mix][source] = None

  def Reset(self):
    for mix, sources in self.mixes.items():
      for source in sources.keys():
        sources[source] = None
        self.UpdateStatus(mix, source)
        
  def ReadStatus(self, qualifier):
    if "Mix" in qualifier and "Source" in qualifier:
      mix = qualifier["Mix"]
      source = qualifier["Source"]

      if mix in self.mixes and source in self.mixes[mix]:
        return self.mixes[mix][source]
    
    return None
    xxx
  def SetStatus(self, mix, source, state):
    mix_mapped = FindKeyByValue(self.mixMap, mix)
    source_mapped  = FindKeyByValue(self.sourceMap, source)

    if mix_mapped and source_mapped:
      self.mixes[mix_mapped][source_mapped]= state.title()
      self.UpdateStatus(mix_mapped, source_mapped)

  def UpdateStatus(self, mix, source):
    if mix in self.mixes and source in self.mixes[mix] and self.name is not None and self.subscribeCallback is not None:
      self.subscribeCallback(self.name, self.mixes[mix][source], { "Mix": mix, "Source": source })

  def IsInitialized(self):
    for mix, sources in self.mixes.items():
      for source, value in sources.items():
        if value is None:
          return False
    
    return True


class AudioMixerInputSourceStatus(AudioMixerInputStatus):
  def __init__(self, listener):
    super().__init__(listener, 'AudioMixerInputSource')

  def GetInitialized(self):
    for mix in self.mixMap.values():
      request = CommandBuilder.GetAudioMixerInputSource(int(mix))

      if request:
        self.listener.Enqueue(request, PRIORITY_QUERY)


class AudioMixerInputVolumeStatus(AudioMixerInputStatus):
  def __init__(self, listener):
    super().__init__(listener, 'AudioMixerInputVolume')

  def SetStatus(self, mix, source, state):
    mix_mapped = FindKeyByValue(self.mixMap, mix)
    source_mapped  = FindKeyByValue(self.sourceMap, source)

    if mix_mapped and source_mapped:
      self.mixes[mix_mapped][source_mapped] = int(state)
      self.UpdateStatus(mix_mapped, source_mapped)

  def GetInitialized(self):
    for mix in self.mixMap.values():
      for source in self.sourceMap.values():
        request = CommandBuilder.GetAudioMixerInputVolume(int(mix), source)

        if request:
          self.listener.Enqueue(request, PRIORITY_QUERY)


class AudioMixerInputMuteStatus(AudioMixerInputStatus):
  def __init__(self, listener):
    super().__init__(listener, 'AudioMixerInputMute')

  def GetInitialized(self):
    for mix in self.mixMap.values():
      for source in self.sourceMap.values():
        request = CommandBuilder.GetAudioMixerInputMute(int(mix), source)

        if request:
          self.listener.Enqueue(request, PRIORITY_QUERY)


class AudioMixerOutputSourceStatus(Status):
  def __init__(self, listener):
    super().__init__('AudioMixerOutputSource')

    self.listener = listener
    self.mixMap = listener.info["AudioMixerOutputMixMap"]
    self.sourceMap = listener.info["AudioMixerOutputSourceMap"]
    
    self.mixes = {}

    for mix in self.mixMap.keys():
      self.mixes[mix] = {}
      for source in self.sourceMap.keys():
        self.mixes[mix][source] = None

  def Reset(self):
    for mix, sources in self.mixes.items():
      for source in sources.keys():
        sources[source] = None
        self.UpdateStatus(mix, source)
        
  def ReadStatus(self, qualifier):
    if "Mix" in qualifier and "Source" in qualifier:
      mix = qualifier["Mix"]
      source = qualifier["Source"]

      if mix in self.mixes and source in self.mixes[mix]:
        return self.mixes[mix][source]
    
    return None
    
  def SetStatus(self, mix, source, state):
    mix_mapped = FindKeyByValue(self.mixMap, mix)
    source_mapped  = FindKeyByValue(self.sourceMap, source)

    if mix_mapped and source_mapped:
      self.mixes[mix_mapped][source_mapped]= state.title()
      self.UpdateStatus(mix_mapped, source_mapped)

  def UpdateStatus(self, mix, source):
    if mix in self.mixes and source in self.mixes[mix] and self.name is not None and self.subscribeCallback is not None:
      self.subscribeCallback(self.name, self.mixes[mix][source], { "Mix": mix, "Source": source })

  def GetInitialized(self):
    for mix in self.mixMap.values():
      request = CommandBuilder.GetAudioMixerOutputSource(int(mix))

      if request:
        self.listener.Enqueue(request, PRIORITY_QUERY)

  def IsInitialized(self):
    for mix, sources in self.mixes.items():
      for source, value in sources.items():
        if value is None:
          return False
    
    return True


class AudioMixerOutputVolumeStatus(Status):
  def __init__(self, listener, name='AudioMixerOutputVolume'):
    super().__init__(name)

    self.listener = listener
    self.mixMap = listener.info["AudioMixerOutputMixMap"]
    self.sourceMap = listener.info["AudioMixerOutputSourceMap"]
    
    self.mixes = {}

    for mix in self.mixMap.keys():
      self.mixes[mix] = None

  def Reset(self):
    for mix in self.mixes.keys():
      self.mixes[mix] = None
      self.UpdateStatus(mix)
        
  def ReadStatus(self, qualifier):
    if "Mix" in qualifier:
      mix = qualifier["Mix"]

      if mix in self.mixes:
        return self.mixes[mix]
    
    return None
    
  def SetStatus(self, mix, state):
    mix_mapped = FindKeyByValue(self.mixMap, mix)

    if mix_mapped:
      self.mixes[mix_mapped] = int(state)
      self.UpdateStatus(mix_mapped)

  def UpdateStatus(self, mix):
    if mix in self.mixes and self.name is not None and self.subscribeCallback is not None:
      self.subscribeCallback(self.name, self.mixes[mix], { "Mix": mix })

  def GetInitialized(self):
    for mix in self.mixMap.values():
      request = CommandBuilder.GetAudioMixerOutputVolume(int(mix))

      if request:
        self.listener.Enqueue(request, PRIORITY_QUERY)

  def IsInitialized(self):
    for mix in self.mixes.values():
      if mix is None:
        return False
    
    return True


class AudioMixerOutputMuteStatus(AudioMixerOutputVolumeStatus):
  def __init__(self, listener):
    super().__init__(listener, name='AudioMixerOutputMute')

  def SetStatus(self, mix, state):
    mix_mapped = FindKeyByValue(self.mixMap, mix)

    if mix_mapped:
      self.mixes[mix_mapped] = state.title()
      self.UpdateStatus(mix_mapped)

  def GetInitialized(self):
    for mix in self.mixMap.values():
      request = CommandBuilder.GetAudioMixerOutputMute(int(mix))

      if request:
        self.listener.Enqueue(request, PRIORITY_QUERY)

################################################################################
## Queues
################################################################################

class Queue():
  def __init__(self):
    self.items = []


  def Enqueue(self, item):
    self.items.insert(0, item)


  def Dequeue(self):
    return self.items.pop()


  def IsEmpty(self):
    return self.items == []


  def Size(self):
    return len(self.items)


  def Clear(self):
    self.items = []


class PriorityQueue():
  def __init__(self, count=None):
    self.count = count if count is not None and count > 0 else 1
    self.queues = [Queue() for i in range(count)]


  def Enqueue(self, item, priority=None):
    if priority is None or priority < 0:
      self.queues[0].Enqueue(item)
    else:
      if self.count > priority:
        self.queues[priority].Enqueue(item)


  def Dequeue(self, priority=None):
    if priority is None or priority < 0:
      for queue in self.queues:
        if queue.Size() > 0:
          return queue.Dequeue()
    
    else:
      if self.count > priority:
        return self.queues[priority].Dequeue()
    
    return None


  def IsEmpty(self, priority=None):
    if priority is None or priority < 0:
      for queue in self.queues:
        if not queue.IsEmpty():
          return False

      return True
    
    else:
      if self.count > priority:
        return self.queues[priority].IsEmpty()
      
      return True


  def Size(self, priority=None):
    count = 0

    if priority is None or priority < 0:
      for queue in self.queues:
        count += queue.Size()

      return count
    
    else:
      if self.count > priority:
        return self.queues[priority].Size()
      
      return count


  def Clear(self, priority=None):
    if priority is None or priority < 0:
      for queue in self.queues:
        queue.Clear()
    
    else:
      if self.count > priority:
        self.queues[priority].Clear()

################################################################################
## Base Interfaces
################################################################################

class SerialClass(SerialInterface, DeviceClass):
  def __init__(self, Host, Port, Baud=9600, Data=8, Parity='None', Stop=1, FlowControl='Off', CharDelay=0, Mode='RS232', Model=None):
    SerialInterface.__init__(self, Host, Port, Baud, Data, Parity, Stop, FlowControl, CharDelay, Mode)
    self.ConnectionType = 'Serial'
    DeviceClass.__init__(self, Model)

    if Model not in self.models: 
      DebugPrint("SerialClass.Constructor", 'Model mismatch')

  def Error(self, message):
    portInfo = 'Host Alias: {0}, Port: {1}'.format(self.Host.DeviceAlias, self.Port)
    DebugPrint("SerialClass.Error", 'Module: {}'.format(__name__), portInfo, 'Error Message: {}'.format(message[0]), sep='\r\n')

  def Discard(self, message):
    self.Error([message])


class SerialOverEthernetClass(EthernetClientInterface, DeviceClass):
  def __init__(self, Hostname, IPPort=23, Protocol='TCP', ServicePort=0, Model=None):
    EthernetClientInterface.__init__(self, Hostname, IPPort, Protocol, ServicePort)
    self.ConnectionType = 'Serial'
    DeviceClass.__init__(self, Model) 

    if Model not in self.models: 
      DebugPrint("SerialOverEthernetClass.Constructor", 'Model mismatch')

  def Error(self, message):
    portInfo = 'IP Address/Host: {0}:{1}'.format(self.Hostname, self.IPPort)
    DebugPrint("SerialOverEthernetClass.Error", 'Module: {}'.format(__name__), portInfo, 'Error Message: {}'.format(message[0]), sep='\r\n')

  def Discard(self, message):
    self.Error([message])

  def Disconnect(self):
    EthernetClientInterface.Disconnect(self)


class EthernetClass(EthernetClientInterface, DeviceClass):
  def __init__(self, Hostname, IPPort=23, Protocol='TCP', ServicePort=0, Model=None):
    EthernetClientInterface.__init__(self, Hostname, IPPort, Protocol, ServicePort)
    self.ConnectionType = 'Ethernet'
    DeviceClass.__init__(self, Model)

    if Model not in self.models: 
      DebugPrint("EthernetClass.Constructor", 'Model mismatch')

  def Error(self, message):
    portInfo = 'IP Address/Host: {0}:{1}'.format(self.Hostname, self.IPPort)
    DebugPrint("EthernetClass.Error", 'Module: {}'.format(__name__), portInfo, 'Error Message: {}'.format(message[0]), sep='\r\n')

  def Discard(self, message):
    self.Error([message])

  def Disconnect(self):
    EthernetClientInterface.Disconnect(self)
